@@ -36,6 +36,13 @@
  * there is no I/O involved and should be accessed only by the main
  * thread */
 
+/* NOTE2: welcome to a really big headache. At the beginning this was
+ * splitted in several classes, one for encoding detection, another
+ * for UTF-8 conversion and another for validation. The reason this is
+ * all together is because we need specific information from all parts
+ * in other to be able to mark characters as invalid if there was some
+ * specific problem on the conversion */
+
 #define GEDIT_DOCUMENT_OUTPUT_STREAM_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE((object),\
 							 GEDIT_TYPE_DOCUMENT_OUTPUT_STREAM,\
 							 GeditDocumentOutputStreamPrivate))
@@ -50,6 +57,9 @@ struct _GeditDocumentOutputStreamPrivate
 	gchar *buffer;
 	gsize buflen;
 
+	gchar *iconv_buffer;
+	gsize iconv_buflen;
+
 	/* Encoding detection */
 	GIConv iconv;
 	GCharsetConverter *charset_conv;
@@ -57,6 +67,9 @@ struct _GeditDocumentOutputStreamPrivate
 	GSList *encodings;
 	GSList *current_encoding;
 
+	gint error_offset;
+	guint n_fallback_errors;
+
 	guint is_utf8 : 1;
 	guint use_first : 1;
 
@@ -131,12 +144,6 @@ gedit_document_output_stream_dispose (GObject *object)
 {
 	GeditDocumentOutputStream *stream = GEDIT_DOCUMENT_OUTPUT_STREAM (object);
 
-	if (stream->priv->iconv != NULL)
-	{
-		g_iconv_close (stream->priv->iconv);
-		stream->priv->iconv = NULL;
-	}
-
 	if (stream->priv->charset_conv != NULL)
 	{
 		g_object_unref (stream->priv->charset_conv);
@@ -152,6 +159,7 @@ gedit_document_output_stream_finalize (GObject *object)
 	GeditDocumentOutputStream *stream = GEDIT_DOCUMENT_OUTPUT_STREAM (object);
 
 	g_free (stream->priv->buffer);
+	g_free (stream->priv->iconv_buffer);
 	g_slist_free (stream->priv->encodings);
 
 	G_OBJECT_CLASS (gedit_document_output_stream_parent_class)->finalize (object);
@@ -177,6 +185,8 @@ gedit_document_output_stream_constructed (GObject *object)
 				      FALSE);
 
 	gtk_source_buffer_end_not_undoable_action (GTK_SOURCE_BUFFER (stream->priv->doc));
+
+	G_OBJECT_CLASS (gedit_document_output_stream_parent_class)->constructed (object);
 }
 
 static void
@@ -219,6 +229,8 @@ gedit_document_output_stream_init (GeditDocumentOutputStream *stream)
 	stream->priv->encodings = NULL;
 	stream->priv->current_encoding = NULL;
 
+	stream->priv->error_offset = -1;
+
 	stream->priv->is_initialized = FALSE;
 	stream->priv->is_closed = FALSE;
 	stream->priv->is_utf8 = FALSE;
@@ -242,7 +254,10 @@ get_encoding (GeditDocumentOutputStream *stream)
 		return (const GeditEncoding *)stream->priv->current_encoding->data;
 	}
 
-	return NULL;
+	stream->priv->use_first = TRUE;
+	stream->priv->current_encoding = stream->priv->encodings;
+
+	return (const GeditEncoding *)stream->priv->current_encoding->data;
 }
 
 static gboolean
@@ -495,77 +510,135 @@ gedit_document_output_stream_get_guessed (GeditDocumentOutputStream *stream)
 guint
 gedit_document_output_stream_get_num_fallbacks (GeditDocumentOutputStream *stream)
 {
-	g_return_val_if_fail (GEDIT_IS_DOCUMENT_OUTPUT_STREAM (stream), FALSE);
+	g_return_val_if_fail (GEDIT_IS_DOCUMENT_OUTPUT_STREAM (stream), 0);
 
-	if (stream->priv->charset_conv == NULL)
-	{
-		return FALSE;
-	}
-
-	return g_charset_converter_get_num_fallbacks (stream->priv->charset_conv) != 0;
+	return stream->priv->n_fallback_errors;
 }
 
-static gboolean
+static void
+apply_error_tag (GeditDocumentOutputStream *stream)
+{
+	GtkTextIter start;
+
+	if (stream->priv->error_offset == -1)
+	{
+		return;
+	}
+
+	gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (stream->priv->doc),
+	                                    &start, stream->priv->error_offset);
+
+	_gedit_document_apply_error_style (stream->priv->doc,
+	                                   &start, &stream->priv->pos);
+
+	stream->priv->error_offset = -1;
+}
+
+static void
+insert_fallback (GeditDocumentOutputStream *stream,
+                 const gchar               *buffer)
+{
+	guint8 out[4];
+	guint8 v;
+	const gchar hex[] = "0123456789ABCDEF";
+
+	/* if we are here is because we are pointing to an invalid char
+	 * so we substitute it by an hex value */
+	v = *(guint8 *)buffer;
+	out[0] = '\\';
+	out[1] = hex[(v & 0xf0) >> 4];
+	out[2] = hex[(v & 0x0f) >> 0];
+	out[3] = '\0';
+
+	gtk_text_buffer_insert (GTK_TEXT_BUFFER (stream->priv->doc),
+	                        &stream->priv->pos, (const gchar *)out, 3);
+
+	++stream->priv->n_fallback_errors;
+}
+
+static void
 validate_and_insert (GeditDocumentOutputStream *stream,
                      const gchar               *buffer,
                      gsize                      count)
 {
-	const gchar *end;
-	gsize nvalid;
-	gboolean valid;
+	GtkTextBuffer *text_buffer;
+	GtkTextIter   *iter;
 	gsize len;
 
+	text_buffer = GTK_TEXT_BUFFER (stream->priv->doc);
+	iter = &stream->priv->pos;
 	len = count;
 
-	/* validate */
-	valid = g_utf8_validate (buffer, len, &end);
-	nvalid = end - buffer;
-
-	if (!valid)
+	while (len != 0)
 	{
-		gsize remainder;
+		const gchar *end;
+		gboolean valid;
+		gsize nvalid;
 
-		remainder = len - nvalid;
-
-		if ((remainder < MAX_UNICHAR_LEN) &&
-		    (g_utf8_get_char_validated (buffer + nvalid, remainder) == (gunichar)-2))
-		{
-			stream->priv->buffer = g_strndup (end, remainder);
-			stream->priv->buflen = remainder;
-			len -= remainder;
-		}
-		else
-		{
-			return FALSE;
-		}
-	}
-	else
-	{
-		gchar *ptr;
+		/* validate */
+		valid = g_utf8_validate (buffer, len, &end);
+		nvalid = end - buffer;
 
 		/* Note: this is a workaround for a 'bug' in GtkTextBuffer where
 		   inserting first a \r and then in a second insert, a \n,
 		   will result in two lines being added instead of a single
 		   one */
 
-		ptr = g_utf8_find_prev_char (buffer, buffer + len);
-
-		if (ptr && *ptr == '\r' && ptr - buffer == len - 1)
+		if (valid)
 		{
-			stream->priv->buffer = g_new (gchar, 1);
-			stream->priv->buffer[0] = '\r';
-			stream->priv->buflen = 1;
+			gchar *ptr;
 
-			--len;
+			ptr = g_utf8_find_prev_char (buffer, buffer + len);
+
+			if (ptr && *ptr == '\r' && ptr - buffer == len - 1)
+			{
+				stream->priv->buffer = g_new (gchar, 1);
+				stream->priv->buffer[0] = '\r';
+				stream->priv->buflen = 1;
+
+				/* Decrease also the len so in the check
+				   nvalid == len we get out of this method */
+				--nvalid;
+				--len;
+			}
 		}
+
+		/* if we've got any valid char we must tag the invalid chars */
+		if (nvalid > 0)
+		{
+			apply_error_tag (stream);
+		}
+
+		gtk_text_buffer_insert (text_buffer, iter, buffer, nvalid);
+
+		/* If we inserted all return */
+		if (nvalid == len)
+		{
+			break;
+		}
+
+		buffer += nvalid;
+		len = len - nvalid;
+
+		if ((len < MAX_UNICHAR_LEN) &&
+		    (g_utf8_get_char_validated (buffer, len) == (gunichar)-2))
+		{
+			stream->priv->buffer = g_strndup (end, len);
+			stream->priv->buflen = len;
+
+			break;
+		}
+
+		/* we need the start of the chunk of invalid chars */
+		if (stream->priv->error_offset == -1)
+		{
+			stream->priv->error_offset = gtk_text_iter_get_offset (&stream->priv->pos);
+		}
+
+		insert_fallback (stream, buffer);
+		++buffer;
+		--len;
 	}
-
-	gtk_text_buffer_insert (GTK_TEXT_BUFFER (stream->priv->doc),
-	                        &stream->priv->pos,
-	                        buffer,
-	                        len);
-
-	return TRUE;
 }
 
 /* If the last char is a newline, remove it from the buffer (otherwise
@@ -607,6 +680,100 @@ end_append_text_to_document (GeditDocumentOutputStream *stream)
 	gtk_source_buffer_end_not_undoable_action (GTK_SOURCE_BUFFER (stream->priv->doc));
 }
 
+static gboolean
+convert_text (GeditDocumentOutputStream *stream,
+              const gchar               *inbuf,
+              gsize                      inbuf_len,
+              gchar                    **outbuf,
+              gsize                     *outbuf_len,
+              GError                   **error)
+{
+	gchar *out, *dest;
+	gsize in_left, out_left, outbuf_size, res;
+	gint errsv;
+	gboolean done, have_error;
+
+	in_left = inbuf_len;
+	/* set an arbitrary length if inbuf_len is 0, this is needed to flush
+	   the iconv data */
+	outbuf_size = (inbuf_len > 0) ? inbuf_len : 100;
+
+	out_left = outbuf_size;
+	out = dest = g_malloc (outbuf_size);
+
+	done = FALSE;
+	have_error = FALSE;
+
+	while (!done && !have_error)
+	{
+		/* If we reached here is because we need to convert the text,
+		   so we convert it using iconv.
+		   See that if inbuf is NULL the data will be flushed */
+		res = g_iconv (stream->priv->iconv,
+		               (gchar **)&inbuf, &in_left,
+		               &out, &out_left);
+
+		/* something went wrong */
+		if (res == (gsize)-1)
+		{
+			errsv = errno;
+
+			switch (errsv)
+			{
+				case EINVAL:
+					/* Incomplete text, do not report an error */
+					stream->priv->iconv_buffer = g_strndup (inbuf, in_left);
+					stream->priv->iconv_buflen = in_left;
+					done = TRUE;
+					break;
+				case E2BIG:
+					{
+						/* allocate more space */
+						gsize used = out - dest;
+
+						outbuf_size *= 2;
+						dest = g_realloc (dest, outbuf_size);
+
+						out = dest + used;
+						out_left = outbuf_size - used;
+					}
+					break;
+				case EILSEQ:
+					/* TODO: we should escape this text.*/
+					g_set_error_literal (error, G_CONVERT_ERROR,
+					                     G_CONVERT_ERROR_ILLEGAL_SEQUENCE,
+					                     _("Invalid byte sequence in conversion input"));
+					have_error = TRUE;
+					break;
+				default:
+					g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_FAILED,
+					             _("Error during conversion: %s"),
+					             g_strerror (errsv));
+					have_error = TRUE;
+					break;
+			}
+		}
+		else
+		{
+			done = TRUE;
+		}
+	}
+
+	if (have_error)
+	{
+		g_free (dest);
+		*outbuf = NULL;
+		*outbuf_len = 0;
+
+		return FALSE;
+	}
+
+	*outbuf = dest;
+	*outbuf_len = out - dest;
+
+	return TRUE;
+}
+
 static gssize
 gedit_document_output_stream_write (GOutputStream            *stream,
 				    const void               *buffer,
@@ -635,14 +802,13 @@ gedit_document_output_stream_write (GOutputStream            *stream,
 		if (ostream->priv->charset_conv == NULL &&
 		    !ostream->priv->is_utf8)
 		{
-			/* FIXME: Add a different domain when we kill gedit_convert */
 			g_set_error_literal (error, GEDIT_DOCUMENT_ERROR,
-					     GEDIT_DOCUMENT_ERROR_ENCODING_AUTO_DETECTION_FAILED,
-					     _("It is not possible to detect the encoding automatically"));
+			                     GEDIT_DOCUMENT_ERROR_ENCODING_AUTO_DETECTION_FAILED,
+			                     _("It is not possible to detect the encoding automatically"));
 			return -1;
 		}
 
-		/* Do not initialize iconv if we are not going to conver anything */
+		/* Do not initialize iconv if we are not going to convert anything */
 		if (!ostream->priv->is_utf8)
 		{
 			gchar *from_charset;
@@ -689,7 +855,7 @@ gedit_document_output_stream_write (GOutputStream            *stream,
 	if (ostream->priv->buflen > 0)
 	{
 		len = ostream->priv->buflen + count;
-		text = g_new (gchar , len + 1);
+		text = g_malloc (len + 1);
 
 		memcpy (text, ostream->priv->buffer, ostream->priv->buflen);
 		memcpy (text + ostream->priv->buflen, buffer, count);
@@ -711,11 +877,11 @@ gedit_document_output_stream_write (GOutputStream            *stream,
 
 	if (!ostream->priv->is_utf8)
 	{
-		gchar *conv_text;
-		gsize conv_read;
-		gsize conv_written;
-		GError *err = NULL;
+		gchar *outbuf;
+		gsize outbuf_len;
 
+		/* check if iconv was correctly initializated, this shouldn't
+		   happen but better be safe */
 		if (ostream->priv->iconv == NULL)
 		{
 			g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_INITIALIZED,
@@ -729,65 +895,57 @@ gedit_document_output_stream_write (GOutputStream            *stream,
 			return -1;
 		}
 
-		/* If we reached here is because we need to convert the text so, we
-		   convert it with the charset converter */
-		conv_text = g_convert_with_iconv (text,
-		                                  len,
-		                                  ostream->priv->iconv,
-		                                  &conv_read,
-		                                  &conv_written,
-		                                  &err);
+		/* manage the previous conversion buffer */
+		if (ostream->priv->iconv_buflen > 0)
+		{
+			gchar *text2;
+			gsize len2;
+
+			len2 = len + ostream->priv->iconv_buflen;
+			text2 = g_malloc (len2 + 1);
+
+			memcpy (text2, ostream->priv->iconv_buffer, ostream->priv->iconv_buflen);
+			memcpy (text2 + ostream->priv->iconv_buflen, text, len);
+
+			text2[len2] = '\0';
+
+			if (freetext)
+			{
+				g_free (text);
+			}
+
+			text = text2;
+			len = len2;
+
+			g_free (ostream->priv->iconv_buffer);
+
+			ostream->priv->iconv_buffer = NULL;
+			ostream->priv->iconv_buflen = 0;
+
+			freetext = TRUE;
+		}
+
+		if (!convert_text (ostream, text, len, &outbuf, &outbuf_len, error))
+		{
+			if (freetext)
+			{
+				g_free (text);
+			}
+
+			return -1;
+		}
 
 		if (freetext)
 		{
 			g_free (text);
 		}
 
-		if (err != NULL)
-		{
-			gsize remainder;
-
-			remainder = len - conv_read;
-
-			/* Store the partial char for the next conversion */
-			if (err->code == G_CONVERT_ERROR_ILLEGAL_SEQUENCE && 
-			    remainder < MAX_UNICHAR_LEN &&
-			    (g_utf8_get_char_validated (text + conv_read, remainder) == (gunichar)-2))
-			{
-				ostream->priv->buffer = g_strndup (text + conv_read, remainder);
-				ostream->priv->buflen = remainder;
-			}
-			else
-			{
-				/* Something went wrong with the conversion,
-				   propagate the error and finish */
-				g_propagate_error (error, err);
-				g_free (conv_text);
-
-				return -1;
-			}
-		}
-
-		text = conv_text;
-		len = conv_written;
-		freetext = TRUE;
+		/* set the converted text as the text to validate */
+		text = outbuf;
+		len = outbuf_len;
 	}
 
-	if (!validate_and_insert (ostream, text, len))
-	{
-		/* TODO: we could escape invalid text and tag it in red
-		 * and make the doc readonly.
-		 */
-		g_set_error (error, G_IO_ERROR,
-		             G_IO_ERROR_INVALID_DATA,
-		             _("Invalid UTF-8 sequence in input"));
-
-		if (freetext)
-		{
-			g_free (text);
-		}
-		return -1;
-	}
+	validate_and_insert (ostream, text, len);
 
 	if (freetext)
 	{
@@ -811,12 +969,44 @@ gedit_document_output_stream_flush (GOutputStream  *stream,
 		return TRUE;
 	}
 
-	if (ostream->priv->buflen == 0)
+	/* if we have converted something flush residual data, validate and insert */
+	if (ostream->priv->iconv != NULL)
 	{
-		return TRUE;
+		gchar *outbuf;
+		gsize outbuf_len;
+
+		convert_text (ostream, NULL, 0, &outbuf, &outbuf_len, error);
+		validate_and_insert (ostream, outbuf, outbuf_len);
+		g_free (outbuf);
+	}
+
+	if (ostream->priv->buflen > 0 && *ostream->priv->buffer != '\r')
+	{
+		/* If we reached here is because the last insertion was a half
+		   correct char, which has to be inserted as fallback */
+		gchar *text;
+
+		if (ostream->priv->error_offset == -1)
+		{
+			ostream->priv->error_offset = gtk_text_iter_get_offset (&ostream->priv->pos);
+		}
+
+		text = ostream->priv->buffer;
+		while (ostream->priv->buflen != 0)
+		{
+			insert_fallback (ostream, text);
+			++text;
+			--ostream->priv->buflen;
+		}
+
+		g_free (ostream->priv->buffer);
+		ostream->priv->buffer = NULL;
 	}
 	else if (ostream->priv->buflen == 1 && *ostream->priv->buffer == '\r')
 	{
+		/* The previous chars can be invalid */
+		apply_error_tag (ostream);
+
 		/* See special case above, flush this */
 		gtk_text_buffer_insert (GTK_TEXT_BUFFER (ostream->priv->doc),
 		                        &ostream->priv->pos,
@@ -826,19 +1016,34 @@ gedit_document_output_stream_flush (GOutputStream  *stream,
 		g_free (ostream->priv->buffer);
 		ostream->priv->buffer = NULL;
 		ostream->priv->buflen = 0;
-
-		return TRUE;
 	}
-	else
+
+	if (ostream->priv->iconv_buflen > 0 )
 	{
-		/* Conversion error */
-		g_set_error (error,
-		             G_IO_ERROR,
-		             G_IO_ERROR_INVALID_DATA,
-		             _("Incomplete UTF-8 sequence in input"));
+		/* If we reached here is because the last insertion was a half
+		   correct char, which has to be inserted as fallback */
+		gchar *text;
 
-		return FALSE;
+		if (ostream->priv->error_offset == -1)
+		{
+			ostream->priv->error_offset = gtk_text_iter_get_offset (&ostream->priv->pos);
+		}
+
+		text = ostream->priv->iconv_buffer;
+		while (ostream->priv->iconv_buflen != 0)
+		{
+			insert_fallback (ostream, text);
+			++text;
+			--ostream->priv->iconv_buflen;
+		}
+
+		g_free (ostream->priv->iconv_buffer);
+		ostream->priv->iconv_buffer = NULL;
 	}
+
+	apply_error_tag (ostream);
+
+	return TRUE;
 }
 
 static gboolean
@@ -851,10 +1056,16 @@ gedit_document_output_stream_close (GOutputStream     *stream,
 	if (!ostream->priv->is_closed && ostream->priv->is_initialized)
 	{
 		end_append_text_to_document (ostream);
+
+		if (ostream->priv->iconv != NULL)
+		{
+			g_iconv_close (ostream->priv->iconv);
+		}
+
 		ostream->priv->is_closed = TRUE;
 	}
 
-	if (ostream->priv->buflen > 0)
+	if (ostream->priv->buflen > 0 || ostream->priv->iconv_buflen > 0)
 	{
 		g_set_error (error,
 		             G_IO_ERROR,
